package org.dict.server;

import java.util.StringTokenizer;
import java.util.Vector;
import org.dict.kernel.*;
import java.io.*;
/**
 * <p>To sort the index file: java org.dict.server.DatabaseFactory indexFile
 *
 * <p>Format of the ini file:
 * <pre>
	# Configuration for ev
	ev.data=anhviet.txt
	ev.index=anhviet.idx
	ev.memoryIndex=true
	ev.comparator=org.dict.vietdict.KeyComparator
	ev.name=English-Vietnamese
	# Configuration for ve
	ve.data=vietanh.txt
	ve.index=vietanh.idx
	ve.memoryIndex=true
	ve.comparator=org.dict.vietdict.KeyComparator
	ev.name=Vietnamese-English
 * </pre>
 */
public class DatabaseFactory {
	static java.util.Hashtable fInstances = new java.util.Hashtable();
/**
 * DatabaseFactory constructor comment.
 */
public DatabaseFactory() {
	super();
}
public static Database createDatabase(
    String id,
    File f,
    java.util.Properties p)
    throws Exception {
    // Create database with data file
    String data = getFilename(f, p.getProperty(id + ".data"));
    log("\nCreating database with data file " + data);
    Database ret = null;
    if (data.toLowerCase().endsWith(".dz")) {
        ret = new org.dict.zip.DictZipDatabase();
    } else {
        ret = new FlatDatabase();
    }
    ret.setID(id);
    ret.setDatafile(data);
    String enc = p.getProperty(id + ".encoding", "utf-8");
    ret.setEncoding(enc);
    // Init key list
    KeyList kl = null;
    String idx = getFilename(f, p.getProperty(id + ".index"));
    boolean inMemory = !"false".equals(p.getProperty(id + ".memoryIndex"));
    if (inMemory) {
        kl = new MemoryKeyList(idx);
    } else {
        kl = new FileKeyList(idx);
    }
    kl.setEncoding(enc);
    ret.setIndex(kl);
    // Init comparator
    IComparator c = new KeyComparator();
    String comp = p.getProperty(id + ".comparator");
    try {
        if (comp != null) {
            c = (IComparator) Class.forName(comp).newInstance();
        }
    } catch (Throwable t) {
        System.out.println("Cannot init comparator: " + t);
    }
    ret.setComparator(c);
    log("Comparator used for keys: " + c);
    // Init morphological functions
    IMorphAnalyzer ma = null;
    String morphName = p.getProperty(id + ".morph");
    try {
        if (morphName != null) {
            ma = (IMorphAnalyzer) Class.forName(morphName).newInstance();
		    log("Morphological functions used: " + ma);
        }
    } catch (Throwable t) {
        System.out.println("No morphological function available: " + t);
    }
    ret.setMorphAnalyzer(ma);
    // Init HTMLPrinter for formatting the output
 	try {
 		initPrinters(id, p);
 	} catch (Exception e) {
 		log(e.toString());
 	}	
    // Search database for its name if necessary
    String name = p.getProperty(id + ".name");
    if (name == null) {
        String dbName = "00-database-short";
        name = ret.define(dbName);
        if (name == null || !name.startsWith(dbName)) {
            name = id;
        } else {
            name = name.substring(dbName.length()).trim();
        }
    }
    ret.setName(name);
    ret.setSourceLanguage(p.getProperty(id + ".sourceLanguage"));
    ret.setTargetLanguage(p.getProperty(id + ".targetLanguage"));
    log("Database created: " + name);
    return ret;
}
static IDictEngine createEngine(String cfg) {
	java.util.Properties p = new java.util.Properties();
	try {
		InputStream in = new FileInputStream(cfg);
		p.load(in);
		in.close();
	} catch (Throwable t) {
		t.printStackTrace();
	}
	File f = new File(cfg);
	log(new java.util.Date().toString());
	log("\nCreate dictionary engine using configuration "+cfg);
	Vector v1 = getDatabaseNames(cfg);
	String[] dbIDs = new String[v1.size()];
	v1.copyInto(dbIDs);
	DictEngine ret = new DictEngine();
	Vector v = new Vector(dbIDs.length);
	for (int i = 0; i<dbIDs.length; i++){
		if ("false".equals(p.getProperty(dbIDs[i]+".use"))) continue;
		try {
			v.addElement(createDatabase(dbIDs[i], f, p));
		} catch (Exception e) {
			log("Cannot create database "+dbIDs[i]);
			log(e.toString());
		}
	}
	IDatabase[] dbs = new IDatabase[v.size()];
	v.copyInto(dbs);
	ret.setDatabases(dbs);
	log("\nDatabase engine created!");
	return ret;
}
public static void dbSort(String in, org.dict.kernel.IComparator c) {
	int last = in.lastIndexOf('.');
	String base = in;
	if (last > 0) {
		base = in.substring(0, last);
	}
	String tmp = base+".tmp";
	String bak = base+".bak";
	System.out.println("Sort file "+in+" and copy original file to "+bak);
	try {
		org.dict.kernel.KeyList kl = new org.dict.kernel.MemoryKeyList(in);
		org.dict.kernel.IKey[] arr = new org.dict.kernel.IKey[kl.size()];
		for (int i = 0; i < arr.length; i++){
			arr[i] = (org.dict.kernel.IKey) kl.get(i);
		}
		org.dict.kernel.ListUtil.sort(arr, c);
		BufferedOutputStream w = new BufferedOutputStream(new FileOutputStream(tmp));
		for (int i = 0; i < arr.length; i++){
			w.write((arr[i].getKey()+"\t"+arr[i].getOffset()+"\t"+arr[i].getLength()).getBytes(kl.getEncoding()));
			if (i<arr.length-1) w.write('\n');
		}
		w.flush();
		w.close();
		System.out.println("File sorted. Rename original file to "+bak);
		new File(in).renameTo(new File(bak));
		System.out.println("Rename temp file to "+in);
		File f = new File(tmp);
		try {
			f.renameTo(new File(in));
		} catch (Exception e) {
			System.out.println("Please copy file "+tmp+" to "+in);
			return;
		}
		if (f.exists()) {
			System.out.println("Delete temp file");
			try {
				f.delete();
			} catch (Exception e) {
				System.out.println("Could not remove temp file "+tmp);
			}
		}
		System.out.println("Done!");
	} catch (Throwable t) {
		t.printStackTrace();
	}
}
static Vector getDatabaseNames(String cfg) {
    Vector v = new Vector(5);
    try {
        BufferedReader r = new BufferedReader(new FileReader(cfg));
        String s;
        while ((s = r.readLine()) != null) {
	        if (s.startsWith("#")) continue;
            int idx = s.indexOf('.');
            if (idx > 0 && s.indexOf('=', idx) > 0) {
                s = s.substring(0, idx);
                if (!v.contains(s)) {
                    v.addElement(s);
                }
            }
        }
        r.close();
    } catch (Throwable t) {
        t.printStackTrace();
    }
    return v;
}
public static IDictEngine getEngine(String cfg) {
	String s = new File(cfg).getAbsolutePath().toLowerCase();
	IDictEngine ret = (IDictEngine) fInstances.get(s);
	if (ret == null) {
		ret = createEngine(cfg);
		fInstances.put(s, ret);
	}
	return ret;
}
static String getFilename(File cfg, String s) {
	File f = new File(s);
	if (f.isAbsolute()) return s;
	return new File(cfg.getParent(), s).getAbsolutePath();
}
static void initPrinters(String id, java.util.Properties p) {

    String formatter;
    // Init printer for formatting HTML output
    AnswerPrinter hp = new DictHTMLPrinter();
    formatter = p.getProperty(id + ".html");
    try {
        if (formatter != null) {
            hp = (HTMLPrinter) Class.forName(formatter).newInstance();
        }
    } catch (Throwable t) {
        System.out.println("Cannot init HTMLPrinter: " + t);
    }
    AnswerPrinter.getPrinters().put(id+"/text/html", hp);
	log("HTMLPrinter used for formatting output: " + hp);

    // Init printer for formatting plain text output
    AnswerPrinter pp = new PlainPrinter();
    formatter = p.getProperty(id + ".txt");
    try {
        if (formatter != null) {
            pp = (PlainPrinter) Class.forName(formatter).newInstance();
        }
    } catch (Throwable t) {
        System.out.println("Cannot init PlainPrinter: " + t);
    }
    AnswerPrinter.getPrinters().put(id+"/text/plain", pp);
	log("PlainPrinter used for formatting output: " + pp);
}
static void log(String msg) {
	Logger.getInstance().log(msg);
}
/**
 * 
 * @param args java.lang.String[]
 */
public static void main(String[] args) {
	if (args.length == 0) {
		System.out.println("Usage: java "+DatabaseFactory.class.getName()+" indexFile");
		System.exit(0);
	}
	dbSort(args[0], new KeyComparator());
}
}
